import xarray as xr
import numpy as np
import glob
import natsort
import vtk
from paraview.simple import *
from vtk.util.numpy_support import vtk_to_numpy
from tqdm import trange
from scipy import ndimage
import tempfile
import glob
import os


def vtk_ascii2binary(adr_disp):
    '''
    Create temporary binary file from a vtk ascii file as there is a bug in reading vtk ascii file.
    '''
    vtk_binary = tempfile.NamedTemporaryFile(delete=False, mode='wb')
    # Create a temporary file
    ascii_file = LegacyVTKReader(registrationName=adr_disp.split('/')[-1], FileNames=[adr_disp])

    # save data
    SaveData(vtk_binary.name+'.vtk', proxy=ascii_file, ChooseArraysToWrite=0,FileType='Binary')


    # Return the path to the temporary file
    return vtk_binary.name+'.vtk'

def vtk2xarray(adr):
    '''
    Convert vtk file to xarray dataset object.
    '''
    reader = vtk.vtkDataSetReader()
    reader.SetFileName(adr)
    reader.Update()

    # Get the output
    ug = reader.GetOutput()
    # Get cell data
    cell_data = ug.GetCellData()

    # Get the number of arrays in the cell data
    num_cell_arrays = cell_data.GetNumberOfArrays()

    vtk_spacing=ug.GetSpacing()
    vtk_dim=ug.GetDimensions()

    ds=xr.Dataset()

    # Iterate over arrays and print their names
    #print("Cell Data:")
    for i in range(num_cell_arrays):
        array_name = cell_data.GetArrayName(i)
        # print("Array {}: {}".format(i, array_name))
        tmp_np=vtk_to_numpy(ug.GetCellData().GetArray(i))
        if len(tmp_np.shape)==1:
            shaped_np=tmp_np.reshape(vtk_dim[1]-1,vtk_dim[0]-1)
            dd=['y','x']
        elif tmp_np.shape[-1]==3:
            shaped_np=tmp_np.reshape(vtk_dim[1]-1,vtk_dim[0]-1,np.shape(tmp_np)[-1])
            shaped_np[:,:,1]=-shaped_np[:,:,1]
            dd=['y','x','d']
            array_name='displacement'
        elif tmp_np.shape[-1]==9:
            shaped_np=tmp_np.reshape(vtk_dim[1]-1,vtk_dim[0]-1,np.shape(tmp_np)[-1])
            shaped_np = shaped_np[..., [0, 4, 8, 1, 2, 5]]
            dd=['y','x','sT']
            if array_name=='e':
                array_name='strain'
        
        ds[array_name]=xr.DataArray(shaped_np,dims=dd)

    ds['x']=np.arange(vtk_dim[0]-1)*vtk_spacing[0]
    ds['y']=(np.arange(vtk_dim[1]-1)*vtk_spacing[1])[::-1]

    return ds


def load_spam1t(adr_disp,adr_strain):
    '''
    load on time step of displacement and strain
    '''
    # convert to binary
    adr_disp_bi=vtk_ascii2binary(adr_disp)
    adr_strain_bi=vtk_ascii2binary(adr_strain)
    # load data
    ds_disp=vtk2xarray(adr_disp_bi)
    ds_strain=vtk2xarray(adr_strain_bi)
    # merge
    ds=ds_disp.merge(ds_strain)

    return ds

def load_spamige(adr_fd,adr_fs,res=1,time_step=1):
    '''
    Load spamige output
    '''

    # read file name from list
    list_d=natsort.natsorted(glob.glob(os.path.join(adr_fd,'*.vtk')))
    list_s=natsort.natsorted(glob.glob(os.path.join(adr_fs,'*.vtk')))

    ds=[]
    
    for i in trange(len(list_d)):
        ds.append(load_spam1t(list_d[i],list_s[i]).expand_dims('time'))

    tds=xr.concat(ds,'time')
    # applied filter
    if 'returnStatus' in list(tds.keys()):
        filter=tds.returnStatus[0,...]!=-5
        filter=ndimage.binary_erosion(filter)
        tds['mask']=xr.DataArray(filter,dims=['y','x'])
        tds=tds.where(tds.mask,drop=True)

    tds['x']=tds['x']*res
    tds['y']=tds['y']*res
    tds['time']=(tds['time']+1)*time_step

    return tds


import spamige.utils as spu
import glob
import os
import natsort
import shutil

def spam_workflow(folder_img,mask,hws=16,iter_max=100,folder_prefix='',keep_only_final=False):
    '''
    workflow designed for 2D correlation from laboratory ice experiment

    :param folder_img: either path to thefolder containing the images or the list of path of each images
    :type folder_img: list or str
    :param hws: spam hws parameter for spam-ldic
    :type hws: int
    :param folder_prefix: folder prefix is wanted
    :type folder_prefix: str
    :param keep_only_final: keep only the correlation between the first and the final image but use the over as intermediate
    :type keep_only_final: bool
    '''


    if isinstance(folder_img, list):
        im_list=folder_img
    else:
        im_list=natsort.natsorted(glob.glob(os.path.join(folder_img,'*.tif*')))


    if len(folder_prefix)!=0:
        if folder_prefix[-1]!='/':
            folder_prefix+='/'
            
        os.makedirs(folder_prefix, exist_ok=keep_only_final)

    for count, im2 in enumerate(im_list[1:]):

        directories_to_remove = []
        
        if count==0:
            # Registration
            res_reg=spu.spamRegistration(im_list[0],im2,mask)
            # Init Guess
            init_guess=glob.glob(os.path.join(res_reg,'*.tsv'))[0]
            os.makedirs(folder_prefix+'spam-init_guess', exist_ok=True)
            shutil.copy(init_guess,folder_prefix+'spam-init_guess/'+str(count)+init_guess.split('/')[-1])
            directories_to_remove.append(res_reg)
            give_ns=True
        else:
            # Init Guess from last ldic
            give_ns=False
            init_guess=natsort.natsorted(glob.glob(os.path.join(folder_prefix+'spam-init_guess/','*.tsv')))[-1]
            print(init_guess)

        # LDIC
        res_ldic=spu.spamLDIC(im_list[0],im2,initGuess=init_guess,mask=mask,hws=hws,give_ns=give_ns,iterations=iter_max)
        directories_to_remove.append(res_ldic)

        # Filtering
        tsvFile=natsort.natsorted(glob.glob(os.path.join(res_ldic,'*.tsv')))
        corrected_ldic,filter_ldic,cor_ldic=spu.spamCorrection(tsvFile[0])
        
        directories_to_remove.append(corrected_ldic)
        directories_to_remove.append(filter_ldic)
        os.makedirs(folder_prefix+'spam-ldic-not-filtered', exist_ok=True)
        if ((count+1)==len(im_list[1:])*keep_only_final)+(not keep_only_final):
            shutil.move(cor_ldic[0].replace('.tsv', '.vtk'),folder_prefix+'spam-ldic-not-filtered/'+cor_ldic[0].replace('.tsv', '.vtk').split('/')[-1])
        os.makedirs(folder_prefix+'spam-ldic-filtered', exist_ok=True)
        ldic_f_vtk=natsort.natsorted(glob.glob(os.path.join(filter_ldic,'*.vtk')))
        if ((count+1)==len(im_list[1:])*keep_only_final)+(not keep_only_final):
            shutil.move(ldic_f_vtk[0],folder_prefix+'spam-ldic-filtered/'+ldic_f_vtk[0].split('/')[-1])

        # Strain rate on filtered data
        strain_f_tsv=natsort.natsorted(glob.glob(os.path.join(filter_ldic,'*.tsv')))
        strain_f=spu.spamStrain(strain_f_tsv[0])
        directories_to_remove.append(strain_f)

        os.makedirs(folder_prefix+'spam-strain-filtered', exist_ok=True)
        strain_f_vtk=natsort.natsorted(glob.glob(os.path.join(strain_f,'*.vtk')))
        if ((count+1)==len(im_list[1:])*keep_only_final)+(not keep_only_final):
            shutil.move(strain_f_vtk[0],folder_prefix+'spam-strain-filtered/'+strain_f_vtk[0].split('/')[-1])

        # Strain rate on NONE filtered data
        strain_nf=spu.spamStrain(cor_ldic[0])
        directories_to_remove.append(strain_nf)

        os.makedirs(folder_prefix+'spam-strain-not-filtered', exist_ok=True)
        strain_nf_vtk=natsort.natsorted(glob.glob(os.path.join(strain_nf,'*.vtk')))
        if ((count+1)==len(im_list[1:])*keep_only_final)+(not keep_only_final):
            shutil.move(strain_nf_vtk[0],folder_prefix+'spam-strain-not-filtered/'+strain_nf_vtk[0].split('/')[-1])

        # Save init guess at each time : can be used for re-run if one day it is implemented
        ldic_f_tsv=natsort.natsorted(glob.glob(os.path.join(filter_ldic,'*.tsv')))
        shutil.move(ldic_f_tsv[0],folder_prefix+'spam-init_guess/'+str(count+1)+ldic_f_tsv[0].split('/')[-1])

        # Remove directories
        for directory in directories_to_remove:
            shutil.rmtree(directory)

    if keep_only_final:
        shutil.rmtree(folder_prefix+'spam-init_guess/')



import glob
import os
import natsort
import shutil
from datetime import datetime
import subprocess


# Wrapper for registration
def spamRegistration( img1, 
                      img2,
                      mask = None, 
                      iterations = 50 ):
    dt = datetime.now()

    prefix =  'tmp_reg'+str(round(datetime.timestamp(dt)))+'/'
    os.mkdir(prefix)
    cmd = [ 'spam-reg', img1, img2, '-bb', '4', '-it', str( iterations ), '-od', prefix, '-mf1', mask ]
    subprocess.Popen(cmd).wait()
    
    return prefix

# Wrapper for LDIC
def spamLDIC( img1,
              img2,
              initGuess = None,
              mask = None,
              iterations = 200,
              hws = 16,
              give_ns=True):
            
    # set ns to 2*hws+1
    ns=2*hws+1

    dt = datetime.now()

    prefix =  'tmp_ldic'+str(round(datetime.timestamp(dt)))+'/'
    os.mkdir(prefix)
    
    cmd = ['spam-ldic', img1, img2, '-pf', initGuess, '-it', str(iterations),'-vtk', '-od', prefix, '-dp', '0.001', '-glt', '1', '-mf1', mask, '-hws', str(hws)]
    
    if give_ns:
        cmd.append('-ns')
        cmd.append(str(ns))
    
    subprocess.Popen(cmd).wait()
    
    return prefix

# Wrapper for correction
def spamCorrection(tsvFile):

    # Correction
    dt = datetime.now()
    OutputCorrected='tmp_corrected'+str(round(datetime.timestamp(dt)))+'/'
    os.mkdir(OutputCorrected)
    
    cmd = ['spam-filterPhiField', '-pf', tsvFile, '-srs', '-srst', '1', '-nn', '4', '-cint', '-F', 'all','-vtk','-od', OutputCorrected]
    subprocess.Popen( cmd ).wait()

    im_cor=natsort.natsorted(glob.glob(os.path.join(OutputCorrected,'*.tsv')))
    
    # Filtering
    dt = datetime.now()
    OutputFiltered = 'tmp_filtered'+str(round(datetime.timestamp(dt)))+'/'
    os.mkdir(OutputFiltered)

    if len(im_cor)==0:
        im_cor=[tsvFile]
        print( '\n Filtering non corrected file !')

    cmd = ['spam-filterPhiField', '-pf', im_cor[0], '-fm', '-fmr', '1', '-F', 'all', 
               '-vtk', '-od', OutputFiltered]
    subprocess.Popen( cmd ).wait()

    #shutil.rmtree(OutputCorrected)

    return OutputCorrected,OutputFiltered,im_cor

# Wrapper for strain calculation
def spamStrain(tsvFile):
    
    dt = datetime.now()
    OutputStrain = 'tmp_strain'+str(round(datetime.timestamp(dt)))+'/'

    cmd = [ 'spam-regularStrain', tsvFile, '-comp', 'vol', 'dev', 'U','e', '-rst', "1", '-vtk','-noTSV', '-od', OutputStrain]
    subprocess.Popen( cmd ).wait()

    return OutputStrain